%% DCS Example with Speckle and DCS classes

addpath("../matlab/")

clear, clc, close all

tic

% Load example speckle series data
load("Speckle_Example_Data.mat")


% Parameters for speckle and DCS classes
pixels_to_bin   = 5;
pixels_to_stack = 10;
sat_threshold   = 1100;
hot_threshold   = 900;
precision       = 1e-3;


% Instantiate Speckle class object
speckle_obj = Speckle(speckle, fps);

% Subtract background
speckle_obj = speckle_obj.clean_time_series(bkg, ...
    Hot_threshold=hot_threshold, ...
    Sat_threshold=sat_threshold, ...
    Subtract_background=1); 

% Sometimes subtracting the background will make some values in the speckle
% series negative. Make them all positive.
speckle_obj = speckle_obj.make_positive();

% Draw raw example pixel time series
figure(), draw_time_series(speckle_obj)


% Bin frames of intensity time series
speckle_obj = speckle_obj.bin_frames(pixels_to_bin);


% Randomly stack pixel intensity time series of the speckle pattern
speckle_obj.speckle_series_ = speckle_obj.speckle_series_(1:1000,:);
speckle_obj = speckle_obj.stack_pixels(pixels_to_stack);

% Draw an example stacked pixel time series
figure(), draw_time_series(speckle_obj), hold off

len = size(speckle_obj.speckle_series_,2);
boundaries  = linspace(length(speckle)/pixels_to_bin, len, pixels_to_stack)/(speckle_obj.framerate_);
hold on, 
xline(boundaries, '--k',LineWidth=0.1)

% Obtain decorrelation time through DCS class and g2 autocorrelation
% The DCS class can output a quick tau_c, mean tau_c and pixelwise tau_c
% measurement:
% * The quick approach assumes the speckle autocorrelation
% decays to 1, which represents no correlation. 
% * The mean approach fits a theory curve to the mean autocorrelation
% trace. It is thus slower but can account for any noise/bias offset S
% that makes the autocorrelation trace not decay to 1 or the type of
% medium.
figure();
DCS_obj     = DCS_MultiExposure(speckle_obj.speckle_series_, speckle_obj.framerate_);
DCS_obj     = DCS_obj.find_g2(Type="fft");

% Fit multi-exponential to mean g2 trace
DCS_obj     = DCS_obj.fit_multiexposure_g2(ExpGrowth=100, ...
                                            EndPoint=size(DCS_obj.mean_g2_autocorrelation_,2)/20, ...
                                            Rho_1=0.1, ...
                                            Tau_1=0.01, ...
                                            Rho_2=0.9, ...
                                            Tau_2=5, ...
                                            Include_offset=0);


%% Plot the g2 curve with fit and tau_c
DCS_obj     = DCS_obj.plot_g2(Errorbars=0);
xline(DCS_obj.fit_myfit_.tau_1, DisplayName="$\tau_1$", color='r', Interpreter='latex')
xline(DCS_obj.fit_myfit_.tau_2, DisplayName="$\tau_2$", color='b', Interpreter='latex')
combined_tau = (DCS_obj.fit_myfit_.rho_1*DCS_obj.fit_myfit_.tau_1 +DCS_obj.fit_myfit_.rho_2*DCS_obj.fit_myfit_.tau_2)/(DCS_obj.fit_myfit_.rho_1+DCS_obj.fit_myfit_.rho_2);
xline(combined_tau, DisplayName="combined $\tau$", color="g", Interpreter='latex')
thesis_fig(gcf);


%% Plot speckle decorrelation time vs position for a single brain slice

load("../Figure4-5/Data.mat")
position = [0 1 2 3 4];

figure()
example_data = squeeze(tau_2(1,:,:));
scatter(position, mean(example_data, 1))
xlim([-0.5 4.5])
ylim([0 7])
xlabel("Position")
ylabel("\tau_c (s)")
thesis_fig(gcf); legend("off")

