classdef DCS_MultiExposure
    % A class that does g2 autocorrelation and related functions for DCS
    % applications.
    %   See DCS_example for how to run class functions.

    properties
        % speckle and g2 inputs and outputs
        sampling_
        sampled_framerate_
        speckle_series_
        speckle_series_pixelwise_corrected_
        g2_autocorrelation_pixelwise_corrected_
        g2_autocorrelation_
        mean_g2_autocorrelation_
        ste_g2_autocorrelation_
        
        % quick method outputs
        tau_c_
        tau_ind_

        % fit outputs
        fit_ci_
        fit_myfit_
        fit_decorrelation_curve_
        fit_fitinfo_
        fit_gof_

        % pixelwise tau_c output
        tau_c_pixelwise_

        % tau_c vs acquisition time output
        tau_c_evolutions_
    end
    
  
    methods

        function obj = DCS_MultiExposure(speckle_series, framerate, options)
            %G2 Construct an instance of this class
            %   Detailed explanation goes here
            arguments
                speckle_series
                framerate
                options.Sampling = 1
            end
            speckle_series  = double(speckle_series);
            obj.sampling_   = options.Sampling;
            obj.sampled_framerate_ = framerate/obj.sampling_;

            % flatten the speckle series to a vector
            if ~ismatrix(speckle_series)
                obj.speckle_series_ = (reshape(speckle_series,...
                    size(speckle_series,1)*size(speckle_series,2),...
                    size(speckle_series,3)));
            else
                obj.speckle_series_ = speckle_series;
            end

            % Variables for g2 estimations
            obj.speckle_series_pixelwise_corrected_     = NaN;
            obj.g2_autocorrelation_pixelwise_corrected_ = NaN;
            obj.g2_autocorrelation_         = NaN;
            obj.mean_g2_autocorrelation_    = NaN;
            obj.ste_g2_autocorrelation_     = NaN;

            obj.tau_ind_            = NaN;
            obj.tau_c_              = NaN;

            obj.fit_myfit_          = NaN;
            obj.fit_decorrelation_curve_ = NaN;
            obj.fit_fitinfo_        = NaN;
            obj.fit_gof_                = NaN;
            
            obj.tau_c_pixelwise_    = NaN;
            obj.tau_c_evolutions_   = NaN;

        end


        function obj = find_g2(obj, options)
            % Calculates the g2 curve and gets a quick estimate for tau_c
            arguments
                obj
                options.Type = "fft" % Type of g2 to use
                options.Precision = 1e-2 % Precision of fast g2 finding method
            end
            disp('Computing g2 trace...')
            switch options.Type
                case {"FFT", "fft", "Fourier", "fourier", "f", "F"}
                    c = fft_g2(obj.speckle_series_(:,1:obj.sampling_:end));
                case {"regular", "r", "R"}
                    c = regular_g2(obj.speckle_series_(:,1:obj.sampling_:end));
                case {"symmetric", "s", "S"}
                    c = symmetric_g2(obj.speckle_series_(:,1:obj.sampling_:end));
                otherwise
                    warning('Unexpected g2 type. Using the fft method...')
                    c = fft_g2(obj.speckle_series_(:,1:obj.sampling_:end));
            end
            fprintf('g2 computation finished. \n');

            % Get pixel autocorrelation traces and mean pixel
            % autocorrelation trace, but first remove rows with nan or inf
            % values.
            [rows,~] = find(any(isnan(c)|isinf(c),2));
            c(rows,:) = [];
            obj.g2_autocorrelation_ = c;
            obj.speckle_series_(rows,:) = [];
            obj.mean_g2_autocorrelation_ = squeeze(mean(c, 1, 'omitnan'));

            % Get standard error for autocorrelation traces.
            autocorrelation_std = squeeze(std(obj.g2_autocorrelation_,1,1,'omitnan'));
            obj.ste_g2_autocorrelation_ = autocorrelation_std./sqrt(size(obj.g2_autocorrelation_,1));

            % Find decorrelation time via quick approach (no fitting)
            % This is where g2 decays to 1+(g2(0)-1)/e
            % First estimate a value for g2(0)
            g20 = mean(obj.mean_g2_autocorrelation_(2:5));
            % Find the tau value where g2 decays to 1+(1/e) its initial value
            tmp_mean_correlation = (obj.mean_g2_autocorrelation_-1-((g20-1)/exp(1)));
            obj.tau_ind_ = find(abs(tmp_mean_correlation(2:end))<options.Precision, 2, "first");
            obj.tau_c_ = mean(obj.tau_ind_*(1/obj.sampled_framerate_));
            fprintf("\tQuick tau_c = %f (s).\n", obj.tau_c_)
        end


        function obj = fit_g2(obj, options)
            % Fits a theory curve to the g2 autocorrelation data.
            arguments
                obj
                options.EndPoint = (length(obj.mean_g2_autocorrelation_)-1)
                options.ExpGrowth = 10
                options.Tau_c = 1 % tau_c estimate for fit
                options.PowerTerm= 0.75 % Power term estimate for fit
                options.Include_offset=0 % include y-offset in fit
            end

            disp("Fitting theory curve to g2 autoccorelation trace...")

            % Define time lags step sizes
            dTau = 1/obj.sampled_framerate_;
            Tau = linspace(0,...
                dTau*length(obj.mean_g2_autocorrelation_),...
                length(obj.mean_g2_autocorrelation_))';
            x = Tau(2:options.EndPoint);
            
            % g2 autocorrelation for fit
            y = obj.mean_g2_autocorrelation_(2:options.EndPoint);
            if ~iscolumn(y) % make column vector if not already
                y = y';
            end
        
            % Beta factor is g2(tau=0) of the g2 curve. 
            % g2(1) is a noise term, so instead use g2(2)
            B2 = obj.mean_g2_autocorrelation_(2)-1; 
            
            % Define a growth factor for exponential weights
            exp_growth = options.ExpGrowth; 
            if exp_growth > 0 % Set weights of fit to be exponential with user defined growth factor
                w = (exp(flip(linspace(0,exp_growth,length(x))))/exp(exp_growth))'; % Exponentially decreasing weights
            else % Equal weighting for each point if desired
                w = ones(length(x),1)'; 
            end
            
            % Fit decaying exponential and find decorrelation time, tau_c
            if options.Include_offset==1
                myfittype = fittype('1 + beta*exp(-(x/tau_c)^pt) + c',...
                    'dependent',{ 'y'},...
                    'independent', {'x'},...
                    'coefficients', {'beta', 'tau_c', 'pt', 'c'});
                [myfit,gof,fitinfo] = fit(x, y, myfittype,...
                    'StartPoint', [B2 options.Tau_c options.PowerTerm 0],...
                    'Lower', [0     0   0   -1],...
                    'Upper', [1e6   1e6 2   1],...
                    'Weight', w);
            else
                myfittype = fittype('1 + beta*exp(-(x/tau_c)^pt)',...
                    'dependent',{ 'y'},...
                    'independent', {'x'},...
                    'coefficients', {'beta', 'tau_c', 'pt'});
                [myfit,gof,fitinfo] = fit(x, y, myfittype,...
                    'StartPoint', [B2 options.Tau_c options.PowerTerm],...
                    'Lower', [0     0   0],...
                    'Upper', [1e6   1e6 2],...
                    'Weight', w);
            end

            obj.fit_ci_     = confint(myfit, 0.95); % 95% confidence intervals       
            obj.fit_myfit_ = myfit;
            obj.fit_decorrelation_curve_ = myfit(x);
            obj.fit_fitinfo_ = fitinfo;
            obj.fit_gof_ = gof;

            disp(myfit)
            fprintf("\tr-squared = %.3f.\n", gof.rsquare)
        end



        function obj = fit_multiexposure_g2(obj, options)
            % Fits a theory curve to the g2 autocorrelation data.
            arguments
                obj
                options.EndPoint = (length(obj.mean_g2_autocorrelation_)-1)
                options.ExpGrowth = 100
                options.Rho_1 = 0.2 % fast tau_c amplitude
                options.Tau_1 = 1 % fast tau_c estimate for fit
                options.Rho_2 = 0.8 % slow tau_c amplitude
                options.Tau_2 = 10 % slow tau_c estimate for fit
                options.Include_offset = 0;
            end

            disp("Fitting multi-exposure theory curve to g2 autoccorelation trace...")

            % Define time lags step sizes
            dTau = 1/obj.sampled_framerate_;
            Tau = linspace(0,...
                dTau*length(obj.mean_g2_autocorrelation_),...
                length(obj.mean_g2_autocorrelation_))';
            x = Tau(2:options.EndPoint);
            
            % g2 autocorrelation for fit
            y = obj.mean_g2_autocorrelation_(2:options.EndPoint);
            if ~iscolumn(y) % make column vector if not already
                y = y';
            end
        
            % Beta factor is g2(tau=0) of the g2 curve. 
            % g2(1) is a noise term, so instead use g2(2)
            B2 = obj.mean_g2_autocorrelation_(2)-1; 
            
            % Define a growth factor for exponential weights
            exp_growth = options.ExpGrowth; 
            if exp_growth > 0 % Set weights of fit to be exponential with user defined growth factor
                w = (exp(flip(linspace(0,exp_growth,length(x))))/exp(exp_growth))'; % Exponentially decreasing weights
            else % Equal weighting for each point if desired
                w = ones(length(x),1)'; 
            end
            
            % Fit multi-exposure exponential and find decorrelation time,
            % tau_c [ https://doi.org/10.1364/BOE.497604 ]
            if options.Include_offset==1
                myfittype = fittype('1 + beta*[rho_1*abs(exp(-(x/tau_1))) + rho_2*abs(exp(-(x/tau_2))) + (1-rho_1-rho_2)^2] + c',...
                    'dependent',{ 'y'},...
                    'independent', {'x'},...
                    'coefficients', {'beta', 'rho_1', 'tau_1', 'rho_2', 'tau_2', 'c'});
                [myfit,gof,fitinfo] = fit(x, y, myfittype,...
                    'StartPoint',   [B2       options.Rho_1     options.Tau_1       options.Rho_2       options.Tau_2,    0],...
                    'Lower',        [B2-0.05  -1                   -1                   -1               0                 -1],...
                    'Upper',        [B2+0.05  1                   50                     1               100               0],...
                    "Weights", w);
            else
                myfittype = fittype('1 + beta*[rho_1*abs(exp(-(x/tau_1))) + rho_2*abs(exp(-(x/tau_2))) + (1-rho_1-rho_2)^2]',...
                    'dependent',{ 'y'},...
                    'independent', {'x'},...
                    'coefficients', {'beta', 'rho_1', 'tau_1', 'rho_2', 'tau_2',});
                [myfit,gof,fitinfo] = fit(x, y, myfittype,...
                    'StartPoint',   [B2       options.Rho_1     options.Tau_1       options.Rho_2     options.Tau_2],...
                    'Lower',        [B2-0.05  0                   0                   0               0            ],...
                    'Upper',        [B2+0.05  1                   10                  1               100          ],...
                    "Weights", w);
            end

            obj.fit_ci_     = confint(myfit, 0.95); % 95% confidence intervals        
            obj.fit_myfit_ = myfit;
            obj.fit_decorrelation_curve_ = myfit(x);
            obj.fit_fitinfo_ = fitinfo;
            obj.fit_gof_ = gof;

            disp(myfit)
            fprintf("\tr-squared = %.3f.\n", gof.rsquare)

        end


        function obj = plot_g2(obj, options)
            arguments
                obj
                options.Errorbars = 0
            end

            disp("Plotting g2 autoccorelation trace and fitted curve...")

            dTau = 1/obj.sampled_framerate_;
            Tau = linspace(0,...
                dTau*length(obj.mean_g2_autocorrelation_),...
                length(obj.mean_g2_autocorrelation_))';
            x = Tau(2:length(obj.fit_decorrelation_curve_)+1);
        
            y = obj.mean_g2_autocorrelation_(2:length(obj.fit_decorrelation_curve_)+1);
            y_err = obj.ste_g2_autocorrelation_(2:length(obj.fit_decorrelation_curve_)+1);
            if ~iscolumn(y) % make column vector if not already
                y = y';
            end
            if ~iscolumn(y_err)
                y = y';
            end
            
            % Make plot
            axis image
            semilogx(x, y, '.k', 'DisplayName', 'Data')
            hold on
            semilogx(x(1:length(obj.fit_decorrelation_curve_)),...
                obj.fit_decorrelation_curve_,...
                'color',[255,99,71 ]/255,... % matplotlib tomato
                'LineWidth',2);
            if options.Errorbars == 1
                hold on
                errorbar(x,y,y_err,'+k',LineWidth=0.01)
                hold off
            end
            yline(((obj.fit_decorrelation_curve_(1)-1)/exp(1)) + 1)
            if ~isempty(obj.tau_c_)
                xline(obj.tau_c_)
            end
            xlabel("$\tau$ (s)",'Interpreter','latex')
            ylabel("g$_2$(~$\tau$)",'Interpreter','latex')
            grid
            yline(1,'--')
            xlim([x(1) x(length(obj.fit_decorrelation_curve_))])
            ylim([0.9 (obj.fit_decorrelation_curve_(1)+0.05)])
            legend({'Data','Exponential Fit'});

           
            grid off
            hold off
        end


        function [obj, outlier_mask] = find_g2_pixelwise(obj, options)
            % Finds g2 for each pixel
            arguments
                obj
                options.Precision = 1E-2 
            end
            disp("Finding tau_c for each pixel...")
            % Get g2(0) value
            g2_zero = mean(obj.g2_autocorrelation_(:,2:5),2);

            % Find frame index where every decorrelation curve decays to (1-g2(0))/e
            tmp_auto_correlation_curve = (obj.g2_autocorrelation_-1-((g2_zero-1)/exp(1)));
            n_pixels = size(obj.g2_autocorrelation_,1);
            tmp_tau_ind = cell(n_pixels,1);
            for ii=1:n_pixels
                tmp_tau_ind{ii} = find(abs(tmp_auto_correlation_curve(ii,2:end))<options.Precision, 1, "first");
            end

            % Get rid of pixels that did not resolve decorrelation time
            empty_rows = cellfun(@isempty,tmp_tau_ind);
            tmp_tau_ind = cell2mat(tmp_tau_ind(~cellfun('isempty',tmp_tau_ind)));
            obj.speckle_series_pixelwise_corrected_ = obj.speckle_series_(~empty_rows,:);
            obj.g2_autocorrelation_pixelwise_corrected_ = obj.g2_autocorrelation_(~empty_rows,:);

            % Convert frame index to decorrelation time for each pixel
            obj.tau_c_pixelwise_ = tmp_tau_ind.*(1/obj.sampled_framerate_);

            % GEt rid of outlier pixels wrt upper and lower quartiles
            outlier_mask = find(isoutlier(obj.tau_c_pixelwise_,"quartiles"));
            obj.tau_c_pixelwise_(outlier_mask,:) = [];
            obj.speckle_series_pixelwise_corrected_(outlier_mask,:) = [];
            obj.g2_autocorrelation_pixelwise_corrected_(outlier_mask,:) = [];
            
            fprintf("\tMean pixelwise tau_c = %f (s).\n", mean(obj.tau_c_pixelwise_))

        end



    % End of methods definitions
    end
% End of class definition    
end


