classdef Speckle
    %SPECKLE A class of operations to perform on speckle pattern before DCS
    %   If wanting to run all the class methods, they should be run in the following 
    %   order:
    %   * clean_time_series
    %   * BW_filter/bin_frames/sliding_window_mean
    %   * stack_pixels
    %   * center_speckle/max_normalise
    %   Note, that not all the methods need to be run, just that if you
    %   want them to, they should be run in the above order for consistent
    %   results. For example, the speckles should not be centered before
    %   stacking, as this prevents one from seeing long decorelation
    %   components in the speckles series.
    
    properties
        raw_speckle_series_
        speckle_series_
        framerate_
        sampling_
        background_
        pixels_to_stack_
        pixels_to_bin_
    end
    
    % Start of methods
    methods

        function obj = Speckle(speckle_series, framerate, options)
            %SPECKLE Construct an instance of this class
            arguments
                speckle_series
                framerate
                options.Sampling = 1
            end
            speckle_series = double(speckle_series);
            obj.raw_speckle_series_ = speckle_series;
            obj.framerate_          = framerate;
            obj.sampling_           = options.Sampling;

            % Reshape speckle series to vector for easier operations
            if ~ismatrix(speckle_series)
                obj.speckle_series_ = reshape(speckle_series, ... 
                    [size(speckle_series,1)*size(speckle_series,2), size(speckle_series,3)]);
            else
                obj.speckle_series_ = speckle_series;
            end
        end


        function [obj, mask] = clean_time_series(obj, bkg, options)
                arguments
                    obj
                    bkg % background
                    options.Sat_threshold = 110 % saturated pixel value
                    options.Hot_threshold = 90 % hot pixel threshold
                    options.Subtract_background = 1 % subtract background
                end

            % Reshape bkg to vector if not already
            if size(bkg,2)>1
                bkg_vec = reshape(bkg, [size(obj.speckle_series_,1), 1]); 
            else
                bkg_vec = bkg;
            end

            % Remove saturated, hot, and dead pixels
            [mask] = sort(cat(1, ...
                        find(max(obj.speckle_series_,[],2)>=options.Sat_threshold), ... % saturated pixels
                        find(bkg_vec>options.Hot_threshold), ... % hot pixels
                        find(mean(obj.speckle_series_,2)==0) ... % dead pixels
                        ));
            if options.Subtract_background == 1
                obj.speckle_series_ = obj.speckle_series_-double(bkg_vec);
                obj.speckle_series_(mask, :) = []; % Omit mask elements from speckle series
            end
        end


        function obj = sliding_window_mean(obj)
            % Moving window mean to sample and clean speckle series along time dimension
            if obj.sampling_ > 1
                obj.speckle_series_ = movmean(obj.speckle_series_, ...
                                            obj.sampling_, ...
                                            2);
            end
        end


        function obj = bin_frames(obj, pixels_to_bin)
            % Bin every pixel_to_bin frames by averaging
            disp("Binning frames...")
            speckle_tmp = obj.speckle_series_;
            n = size(speckle_tmp,1);
            m = size(speckle_tmp,2);
            obj.speckle_series_ = squeeze(mean(reshape(speckle_tmp,[n pixels_to_bin, m/pixels_to_bin]), 2));
            obj.framerate_ = obj.framerate_/pixels_to_bin;

            % This should output 1 if reshape is correct
            disp(isequal(mean(speckle_tmp(1,1:pixels_to_bin) ), obj.speckle_series_(1,1) ))
            
            obj.pixels_to_bin_ = pixels_to_bin;
            disp("Finished binning frames.")
        end


        function obj = make_positive(obj)
            obj.speckle_series_ = obj.speckle_series_ + abs(min(obj.speckle_series_, 2));
        end


        function obj = BW_filter(obj, options)  
            % BW_filter Butterworth filter the pixel time series
            arguments
                obj
                options.Order = 1
                options.Wn = 5/obj.framerate_
                options.ftype = "low"
            end

            [b, a] = butter(options.Order, options.Wn, options.ftype);
            obj.speckle_series_ = filter(b, a, obj.speckle_series_, [], 2);

            % Omit some frames due to boundary conditions of BW filter
            skipped_frames = length(obj.speckle_series_)*0.002;
            obj.speckle_series_ = obj.speckle_series_(:,skipped_frames:end-skipped_frames);
        end


        function obj = stack_pixels(obj, pixels_to_stack)
        %stack_pixels Stack the time series of pixels. This will
        %effectively increase the length of the time series by a factor
        %of the how many pixels are stacked, and has useful
        %implications for finding the correct decorrelation time.

        % Warning: only call this function last.

        if pixels_to_stack>1
            fprintf("Randomly stacking %d pixels... ", pixels_to_stack)

            obj.pixels_to_stack_ = pixels_to_stack;
            pixels = size(obj.speckle_series_,1);
            frames = size(obj.speckle_series_,2);
            
            %   First check that total pixels are divisible by number of stacks
            if mod(pixels,obj.pixels_to_stack_)==0
                prev_speckle_series        = obj.speckle_series_;
                new_pixels = pixels/obj.pixels_to_stack_;
                stacked_speckle_series     = zeros(new_pixels,frames*obj.pixels_to_stack_);
                for ii=1:new_pixels
                    %   get size of previous speckle in iteration
                    msize   = size(prev_speckle_series,1);
                    %   get no_stacks random indices of that speckle
                    stacking_index  = randsample(msize,obj.pixels_to_stack_);
                    %   Create iith stacked pixel
                    stacked_speckle_series(ii,:)   = reshape(prev_speckle_series(stacking_index,:)',[1,frames*obj.pixels_to_stack_]);
                    %   remove indices that were used in this iteration
                    prev_speckle_series(stacking_index,:) = [];
                end
                obj.speckle_series_ = stacked_speckle_series;
                disp("Done stacking.")
            else
                error("Please enter a number of stacks that can divide total number of pixels.")
            end
            
        end    
    end


    function draw_time_series(obj)
        frames = size(obj.speckle_series_,2);
        x = linspace(0,frames/obj.framerate_,frames);
        y = squeeze(obj.speckle_series_(round(end/2),:));
        plot(x,y)
        xlabel("Time (s)"), ylabel("Intensity")
        thesis_fig(gcf); legend("off")
    end


    end
    % End of methods
end


% Fri 28/07/23: Check that this class works.

