function g2 = g2_wip(I)
% Author: Phil Binner, p.binner.1@research.gla.ac.uk
% Computes g2 using a vectorised FFT method.
% I is an input speckle series. It can be 3D (x pixel, y pixel, frames) or
% 2D (x pixel, frames).

    if gpuDeviceCount > 0 % Run this block if GPU is detected.

        if ndims(I) == 3 % Run this block if input is 3D (x,y,time).

            % Convert speckle series to GPU array.
            I = gpuArray(I); 

            % The following will be used later to normalise the g2
            len = size(I,3); % Size of time dimension.
            frame_norm = 1./(len-((1:(len))-1)); % This array will be used to find the mean intensity of each pixel time series.
            frame_norm = repmat(frame_norm.',[1,size(I,1),size(I,2)]); % Repeat array to 3D
            frame_norm = permute(frame_norm, [2 3 1]); % Permute array to correct order.
            intensity_norm = mean(abs(I),3).^2;

            % Zero-pad the intensity time series
            % The pad size should be such that  the total signal length > 2*initial signal
            pad_size = ceil(size(I,3)/2 )+5; 
            I_padded = padarray(I,[0,0,pad_size],0,'both');
            clear I

            % Find g2
            % - unnormalised g2 = IFT{[FT{g(x)} .* FT{g*(-x)}]}
            g2 = real(ifftshift(ifft(fft(fftshift(I_padded ),[],3 ) .* fft(fftshift(flip(I_padded,3 ) ),[],3) ,[],3) ) );
            clear I_padded
            g2 = g2(:,:,(1+ceil(end/2)):(end-5)); % Crop out the extra time dimension length from zero padding

            % Now normalise the g2 by mean intensity and number of frames 
            % and then convert back from GPU array
            g2 = gather( (g2.*frame_norm)./intensity_norm );
        

        elseif ismatrix(I) % Run this block if input is 2D (x,time).

            % Convert speckle series to GPU array.
            I = gpuArray(I); 

            % The following will be used later to normalise the g2
            len = size(I,2);
            frame_norm = (1./((len)-((1:(len))-1))); 
            intensity_norm = mean(abs(I),2).^2;

            % the pad size should be such that  the total signal length > 2*initial signal
            pad_size = ceil(size(I,2)/2 )+5; 
            I_padded = padarray(I,[0,pad_size],0,'both');
            clear I
            
            % Find g2
            % - unnormalised g2 = IFT{[FT{g(x)} .* FT{g*(-x)}]}
            g2 = real(ifftshift(ifft(fft(fftshift(I_padded ),[],2 ) .* fft(fftshift(fliplr(I_padded ) ),[],2) ,[],2) ) );
            clear I_padded

            g2 = g2(:,(1+ceil(end/2)):(end-5));
            
            % Now normalise the g2 by mean intensity and number of frames 
            % and then convert back from GPU array
            g2 = gather((g2.*frame_norm)./intensity_norm); 
        end


    % If there is no GPU available, run below. The code is the exact same
    % save for gpuArray() and gather() functions.
    else % Run this block if no GPU is detected.
            if ndims(I) == 3 
            len = size(I,3);
            frame_norm = 1./(len-((1:(len))-1)); 
            frame_norm = repmat(frame_norm.',[1,size(I,1),size(I,2)]); 
            frame_norm = permute(frame_norm, [2 3 1]); 
            intensity_norm = mean(abs(I),3).^2;
            pad_size = ceil(size(I,3)/2 )+5; 
            I_padded = padarray(I,[0,0,pad_size],0,'both');
            clear I
            g2 = real(ifftshift(ifft(fft(fftshift(I_padded ),[],3 ) .* fft(fftshift(flip(I_padded,3 ) ),[],3) ,[],3) ) );
            clear I_padded
            g2 = g2(:,:,(1+ceil(end/2)):(end-5));
            g2 = (g2.*frame_norm)./intensity_norm;
        
            elseif ismatrix(I)
                len = size(I,2);
                frame_norm = (1./((len)-((1:(len))-1))); 
                intensity_norm = mean(abs(I),2).^2;
                pad_size = ceil(size(I,2)/2 )+5; 
                I_padded = padarray(I,[0,pad_size],0,'both');
                clear I
                g2 = real(ifftshift(ifft(fft(fftshift(I_padded ),[],2 ) .* fft(fftshift(fliplr(I_padded ) ),[],2) ,[],2) ) );
                clear I_padded
                g2 = g2(:,(1+ceil(end/2)):(end-5));
                g2 = (g2.*frame_norm)./intensity_norm; 
            end     

    end
end
